#pragma once
#include <dspsim/context.h>
#include <dspsim/model.h>
#include <dspsim/signal.h>
#include <dspsim/clock.h>
#include <dspsim/port.h>
#include <dspsim/axis.h>

#include <string>

namespace dspsim
{
    // std::string hello_from_bin() { return "DSPSim!"; }
    std::string hello_from_bin();
} // namespace dspsim
